/**
 * ANTLR3 grammar for LIN Description File (LDF) format, LIN version 2.2.<br>
 * The following files are generated by antlr: {@link: LINLexer}, {@link: LINParser}
 * Author: Philip Stoehrer
 */
grammar LDF;

options {
    output=AST; // AST = Abstract Syntax Tree
    k=4;        // lookahead
}
tokens {    

    LinDescriptionFileDefKey    = 'LIN_description_file';

    LinProtocolVersionDefKey    = 'LIN_protocol_version';

    LinLanguageVersionDefKey    = 'LIN_language_version';

    LinSpeedDefKey              = 'LIN_speed';

    ChannelNameDefKey           = 'Channel_name';

    NodesDefKey                 = 'Nodes';
    MasterDefKey                = 'Master';
    SlavesDefKey                = 'Slaves';
    CompositeDefKey             = 'composite';
    ConfigurationKey            = 'configuration';

    SignalDefKey                = 'Signals';

    DiagSignalDefKey            = 'Diagnostic_signals';
    MasterReqB0                 = 'MasterReqB0';
    MasterReqB1                 = 'MasterReqB1';
    MasterReqB2                 = 'MasterReqB2';
    MasterReqB3                 = 'MasterReqB3';
    MasterReqB4                 = 'MasterReqB4';
    MasterReqB5                 = 'MasterReqB5';
    MasterReqB6                 = 'MasterReqB6';
    MasterReqB7                 = 'MasterReqB7';
    SlaveRespB0                 = 'SlaveRespB0';
    SlaveRespB1                 = 'SlaveRespB1';
    SlaveRespB2                 = 'SlaveRespB2';
    SlaveRespB3                 = 'SlaveRespB3';
    SlaveRespB4                 = 'SlaveRespB4';
    SlaveRespB5                 = 'SlaveRespB5';
    SlaveRespB6                 = 'SlaveRespB6';
    SlaveRespB7                 = 'SlaveRespB7';

    FrameDefKey                 = 'Frames';

    SporadicFrameDefKey         = 'Sporadic_frames';

    EventTriggeredFrameDefKey   = 'Event_triggered_frames';

    DiagFrameDefKey             = 'Diagnostic_frames';
    MasterReq                   = 'MasterReq';
    SlaveResp                   = 'SlaveResp';

    NodeAttributesDefKey        = 'Node_attributes';
    LinProtocolKey              = 'LIN_protocol';
    ConfiguredNADKey            = 'configured_NAD';
    InitialNADKey               = 'initial_NAD';
    ProductIDKey                = 'product_id';
    ResponseErrorKey            = 'response_error';
    FaultStateSignalsKey        = 'fault_state_signals';
    P2MinKey                    = 'P2_min';
    StMinKey                    = 'ST_min';
    NAsTimeoutKey               = 'N_As_timeout';
    NCrTimeoutKey               = 'N_Cr_timeout';
    ConfigurableFramesKey       = 'configurable_frames';

    ScheduleTableDefKey         = 'Schedule_tables';
    AssignNAD                   = 'AssignNAD';
    ConditionalChangeNAD        = 'ConditionalChangeNAD';
    DataDump                    = 'DataDump';
    SaveConfiguration           = 'SaveConfiguration';
    AssignFrameIdRange          = 'AssignFrameIdRange';
    FreeFormat                  = 'FreeFormat';
    AssignFrameId               = 'AssignFrameId';

    SignalGroupsDefKey          = 'Signal_groups';

    SignalEncodingTypeDefKey    = 'Signal_encoding_types';
    LogicalValue                = 'logical_value';
    PhysicalValue               = 'physical_value';
    BcdValue                    = 'bcd_value';
    AsciiValue                  = 'ascii_value';

    SignalRepresentationDefKey  = 'Signal_representation';

    Subscribers;
    Array;
}

@header {
/**
 * Copyright (C) 2012 Telemotive AG.
 * All rights reserved.
 */
package de.telemotive.bpng.client.lin.ldfparser.grammar;
}

@lexer::header{
/**
 * Copyright (C) 2012 Telemotive AG.
 * All rights reserved.
 */
package de.telemotive.bpng.client.lin.ldfparser.grammar;
}

@members {
    public static String createTreeString(Tree tree) {
        return createTreeString(tree, 0).toString();
    }

    private static StringBuilder createTreeString(Tree t, int indent) {
        StringBuilder result = new StringBuilder();
        if (t != null) {
            StringBuffer sb = new StringBuffer(indent);
            for (int i = 0; i < indent; i++) {
                sb = sb.append("   ");
            }
            result.append(sb.toString());
            result.append(tokenNames[t.getType()]);
            result.append(": ");
            result.append(t.getText());
            result.append("\n");

            for (int i = 0; i < t.getChildCount(); i++) {
                result.append(createTreeString((CommonTree) t.getChild(i), indent + 1));
            }
        }
        return result;
    }
}

start:  linDescriptionFileDef
        linProtocolVersionDef
        linLanguageVersionDef
        linSpeedDef
        channelNameDef?
        nodeDef
        nodeCompositionDef?
        signalsDef
        diagSignalDef?
        framesDef
        sporadicFramesDef?
        eventTriggeredFramesDef?
        diagFramesDef?
        nodeAttributesDef
        scheduleTablesDef
        signalGroupsDef?
        signalEncodingTypesDef?
        signalRepresentationsDef?;

linDescriptionFileDef: LinDescriptionFileDefKey ';'!;

linProtocolVersionDef: LinProtocolVersionDefKey '='! STRING ';'!;

linLanguageVersionDef: LinLanguageVersionDefKey '='! STRING ';'!;

linSpeedDef: LinSpeedDefKey '='! real 'kbps'! ';'!;

channelNameDef: ChannelNameDefKey '='! STRING ';'!;

nodeDef: NodesDefKey^ '{'! masterDef slavesDef '}'!;
masterDef: MasterDefKey^ ':'! IDENTIFIER ','! real 'ms'! ','! real 'ms'! ';'!;
slavesDef: SlavesDefKey^ ':'! IDENTIFIER (','! IDENTIFIER)* ';'!;

nodeCompositionDef: CompositeDefKey^ '{'! configuration+ '}'!;
configuration: ConfigurationKey^ IDENTIFIER '{'! compositeNode+ '}'!;
compositeNode: IDENTIFIER^ '{'! IDENTIFIER (','! IDENTIFIER)* ';'! '}'!;

signalsDef: SignalDefKey^ '{'! signalDef+ '}'!;
signalDef: IDENTIFIER^ ':'! integer ','! initValue ','! IDENTIFIER (','! subscribers)? ';'!;
initValue: integer | ('{' integer (',' integer)* '}' -> ^(Array integer+));
subscribers: IDENTIFIER (',' IDENTIFIER)* -> ^(Subscribers IDENTIFIER+);

diagSignalDef: DiagSignalDefKey^ '{'! masterReqB0 masterReqB1 masterReqB2 masterReqB3 masterReqB4 masterReqB5 masterReqB6 masterReqB7 slaveRespB0 slaveRespB1 slaveRespB2 slaveRespB3 slaveRespB4 slaveRespB5 slaveRespB6 slaveRespB7 '}'!;
masterReqB0: MasterReqB0^ ':'! integer ','! integer ';'!;
masterReqB1: MasterReqB1^ ':'! integer ','! integer ';'!;
masterReqB2: MasterReqB2^ ':'! integer ','! integer ';'!;
masterReqB3: MasterReqB3^ ':'! integer ','! integer ';'!;
masterReqB4: MasterReqB4^ ':'! integer ','! integer ';'!;
masterReqB5: MasterReqB5^ ':'! integer ','! integer ';'!;
masterReqB6: MasterReqB6^ ':'! integer ','! integer ';'!;
masterReqB7: MasterReqB7^ ':'! integer ','! integer ';'!;
slaveRespB0: SlaveRespB0^ ':'! integer ','! integer ';'!;
slaveRespB1: SlaveRespB1^ ':'! integer ','! integer ';'!;
slaveRespB2: SlaveRespB2^ ':'! integer ','! integer ';'!;
slaveRespB3: SlaveRespB3^ ':'! integer ','! integer ';'!;
slaveRespB4: SlaveRespB4^ ':'! integer ','! integer ';'!;
slaveRespB5: SlaveRespB5^ ':'! integer ','! integer ';'!;
slaveRespB6: SlaveRespB6^ ':'! integer ','! integer ';'!;
slaveRespB7: SlaveRespB7^ ':'! integer ','! integer ';'!;

framesDef: FrameDefKey^ '{'! frameDef+ '}'!;
frameDef: IDENTIFIER^ ':'! integer ','! IDENTIFIER ','! integer '{'! frameSignalDef+ '}'!;
frameSignalDef: IDENTIFIER^ ','! integer ';'!;

sporadicFramesDef: SporadicFrameDefKey^ '{'! sporadicFrameDef+ '}'!;
sporadicFrameDef: IDENTIFIER^ ':'! IDENTIFIER(','! IDENTIFIER)* ';'!;

eventTriggeredFramesDef: EventTriggeredFrameDefKey^ '{'! eventTriggeredFrameDef+ '}'!;
eventTriggeredFrameDef: IDENTIFIER^ ':'! IDENTIFIER ','! integer (','! IDENTIFIER)+ ';'!;

diagFramesDef: DiagFrameDefKey^ '{'! diagFrameMasterReq diagFrameSlaveResp '}'!;
diagFrameMasterReq: MasterReq^ ':'! integer '{'! diagFrameMasterReqB0 diagFrameMasterReqB1
                            diagFrameMasterReqB2 diagFrameMasterReqB3
                            diagFrameMasterReqB4 diagFrameMasterReqB5
                            diagFrameMasterReqB6 diagFrameMasterReqB7 '}'!;
diagFrameMasterReqB0: MasterReqB0^ ','! integer ';'!;
diagFrameMasterReqB1: MasterReqB1^ ','! integer ';'!;
diagFrameMasterReqB2: MasterReqB2^ ','! integer ';'!;
diagFrameMasterReqB3: MasterReqB3^ ','! integer ';'!;
diagFrameMasterReqB4: MasterReqB4^ ','! integer ';'!;
diagFrameMasterReqB5: MasterReqB5^ ','! integer ';'!;
diagFrameMasterReqB6: MasterReqB6^ ','! integer ';'!;
diagFrameMasterReqB7: MasterReqB7^ ','! integer ';'!;
diagFrameSlaveResp: SlaveResp^ ':'! integer '{'! diagFrameSlaveRespB0 diagFrameSlaveRespB1
                                    diagFrameSlaveRespB2 diagFrameSlaveRespB3
                                    diagFrameSlaveRespB4 diagFrameSlaveRespB5
                                    diagFrameSlaveRespB6 diagFrameSlaveRespB7 '}'!;
diagFrameSlaveRespB0: SlaveRespB0^ ','! integer ';'!;
diagFrameSlaveRespB1: SlaveRespB1^ ','! integer ';'!;
diagFrameSlaveRespB2: SlaveRespB2^ ','! integer ';'!;
diagFrameSlaveRespB3: SlaveRespB3^ ','! integer ';'!;
diagFrameSlaveRespB4: SlaveRespB4^ ','! integer ';'!;
diagFrameSlaveRespB5: SlaveRespB5^ ','! integer ';'!;
diagFrameSlaveRespB6: SlaveRespB6^ ','! integer ';'!;
diagFrameSlaveRespB7: SlaveRespB7^ ','! integer ';'!;

nodeAttributesDef: NodeAttributesDefKey^ '{'! nodeAttributeDef+ '}'!;
nodeAttributeDef: IDENTIFIER^ '{'! linProtocol configuredNAD initialNAD? attributesDef '}'!;
linProtocol: LinProtocolKey^ '='! STRING ';'!;
configuredNAD: ConfiguredNADKey^ '='! integer ';'!;
initialNAD: InitialNADKey^ '='! integer ';'!;
attributesDef: productID responseError faultStateSignals? p2Min? stMin? nAsTimeout? nCrTimeout? (configurableFrames20Def | configurableFrames21Def);
productID: ProductIDKey^ '='! integer ','! integer (','! integer)* ';'!;
responseError: ResponseErrorKey^ '='! IDENTIFIER ';'!;
faultStateSignals: FaultStateSignalsKey^ '='! IDENTIFIER ';'!;
p2Min: P2MinKey^ '='! real 'ms'! ';'!;
stMin: StMinKey^ '='! real 'ms'! ';'!;
nAsTimeout: NAsTimeoutKey^ '=' real 'ms'! ';'!;
nCrTimeout: NCrTimeoutKey^ '=' real 'ms'! ';'!;
configurableFrames20Def: ConfigurableFramesKey^ '{'! configurableFrame20Def+ '}'!;
configurableFrame20Def: IDENTIFIER^ '='! integer ';'!;
configurableFrames21Def: ConfigurableFramesKey^ '{'! configurableFrame21Def+ '}'!;
configurableFrame21Def: IDENTIFIER ';'!;

scheduleTablesDef: ScheduleTableDefKey^ '{'! scheduleTableDef+ '}'!;
scheduleTableDef: IDENTIFIER^ '{'! scheduleTableRowDef+ '}'!;
scheduleTableRowDef : commandDef^ 'delay' real 'ms'! ';'!;
commandDef: MasterReq^
                | SlaveResp^
                | AssignNAD^ '{'! IDENTIFIER '}'!
                | ConditionalChangeNAD^ '{'! nad ','! id ','! nadByte ','! mask ','! inv ','! newNAD '}'!
                | DataDump^ '{'! IDENTIFIER ','! d1 ','! d2 ','! d3 ','! d4 ','! d5 '}'!
                | SaveConfiguration^ '{'! IDENTIFIER '}'!
                | AssignFrameIdRange^ '{'! IDENTIFIER ','! integer (','! integer ','! integer ','! integer ','! integer)?'}'!
                | FreeFormat^ '{'! d1 ','! d2 ','! d3 ','! d4 ','! d5 ','! d6 ','! d7 ','! d8 '}'!
                | AssignFrameId^ '{'! IDENTIFIER ','! IDENTIFIER '}'!
                | IDENTIFIER^;
nad: integer;
id: integer;
nadByte: integer;
mask: integer;
inv: integer;
newNAD: integer;
d1: integer;
d2: integer;
d3: integer;
d4: integer;
d5: integer;
d6:integer;
d7: integer;
d8: integer;

signalGroupsDef: SignalGroupsDefKey^ '{'! signalGroupDef+ '}'!;
signalGroupDef: IDENTIFIER^ ':'! integer '{'! signalGroupRowDef+ '}'!;
signalGroupRowDef: IDENTIFIER^ ','! integer ';'!;

signalEncodingTypesDef: SignalEncodingTypeDefKey^ '{'! signalEncodingTypeDef+ '}'!;
signalEncodingTypeDef: IDENTIFIER^ '{'! (logicalValue | physicalRange | bcdValue | asciiValue )+ '}'!;
logicalValue: LogicalValue^ ','! integer (','! STRING)? ';'!;
physicalRange: PhysicalValue^ ','! integer ','! integer ','! real ','! real (','! STRING)? ';'!;
bcdValue: BcdValue ';'!;
asciiValue: AsciiValue ';'!;

signalRepresentationsDef: SignalRepresentationDefKey^ '{'! signalRepresentationDef+ '}'!;
signalRepresentationDef: IDENTIFIER^ ':'! IDENTIFIER (','! IDENTIFIER)* ';'!;

integer: INTEGER | HEX_NUMBER;
real: INTEGER | DECIMAL;

MULTILINE_COMMENT:   '/*' (options {greedy=false;} : .)* '*/' {skip();};
SINGLELINE_COMMENT: '//' ( (~('\r'|'\n'))* ) NEWLINE {skip();};
STRING: '"' ~('"')* '"';
WS: WS_CHARACTER+ {skip();};

INTEGER: ('-')? DIGIT+;
DECIMAL: ('-')? DIGIT+ '.' DIGIT* EXPONENT? ;
HEX_NUMBER: '0' ('x'|'X') HEX_DIGIT+;
IDENTIFIER: (LETTER|'_') ALPHA_NUMERIC*;

fragment ALPHA_NUMERIC: (LETTER|DIGIT|'_');
fragment NEWLINE: '\r'? '\n'; // Supporting Windows and Unix linebreaks. (no Macintosh).
fragment WS_CHARACTER: ' '|'\t'|'\r'|'\n';
fragment DIGIT: '0'..'9';
fragment UPPER_CASE_LETTER: 'a'..'z';
fragment LOWER_CASE_LETTER: 'A'..'Z';
fragment LETTER: UPPER_CASE_LETTER|LOWER_CASE_LETTER;
fragment EXPONENT: ('e'|'E') ('-'|'+')? DIGIT+;
fragment HEX_DIGIT: '0'..'9' | 'A'..'F' | 'a'..'f';
